###########################################
#             Grasso.py                   #
###########################################
#Author: Chris Fenton - 11/17/24

import cv2
from llama_cpp import LlamaCache
from llama_cpp import Llama
from llama_cpp.llama_chat_format import Llava15ChatHandler
from llama_cpp.llama_types import *
import base64
import time
import random
import serial
import sys
from datetime import datetime
from datetime import timedelta
from datetime import date

def image_to_base64_data_uri(file_path):
    with open(file_path, "rb") as img_file:
        base64_data = base64.b64encode(img_file.read()).decode('utf-8')
        return f"data:image/png;base64,{base64_data}"

#Load vision model - INSERT YOUR MODEL HERE
view_mmproj = "./models/llava-v1.6-7b-mmproj.gguf"
view_model  = "./models/llava-1.6-mistral-7b.Q8_0.gguf" 

#Load brain model - INSERT YOUR MODEL HERE
brain_model = "./models/mistral-7b-instruct-v0.1.Q8_0.gguf"

brain_cache = LlamaCache()
view_cache  = LlamaCache()

chat_handler = Llava15ChatHandler(clip_model_path=view_mmproj)

#Text-to-speech synthesizer
use_voice = True 
voice = serial.Serial('/dev/ttyUSB0', 9600, timeout=1)

#Arduino-based Motor controller
for tgt in ['/dev/ttyACM0','/dev/ttyACM1','/dev/ttyACM2']:
    try:
        motor = serial.Serial(tgt, 9600, timeout=10)
        break
    except Exception:
        pass

#Configure Accent SA voice box
#Send <ESC>P3 <ESC>V4
cmd1='P1'   #P9/V9 is an okay combo
cmd2='V9'   #1 is pretty good
voice.write(chr(27).encode())
voice.write(cmd1.encode())
voice.write(chr(27).encode())
voice.write(cmd2.encode())

awake = 'Grasso! Is! Alive!'

if use_voice:
    voice.write(awake.encode())

logfilename = './grasso_logs/' + datetime.now().strftime('grasso_%Y_%m_%d_%H_%M.log')
logfile = open(logfilename, 'w')

brain_llm = Llama( model_path=brain_model, verbose=False, n_ctx=2048, set_cache=brain_cache )

#Webcam
videoStreamObject = cv2.VideoCapture(0)

count       = 0
valid_count = 0

view_llm = Llama(model_path=view_model,
            chat_handler=chat_handler,
            chat_format='chatml',
            stop=["\n", "###"],
            n_ctx=2048,
            logits_all=True,
            verbose=False
            )

"""
"[INSTR] You are a stationary, solar-powered robot named Grasso, who lives in a yard. Your current battery level is full. You are feeling happy.\n
You observe the following scene:  The image depicts a cozy living room with a blue couch and a chair placed in front of it, creating a comfortable seating area. A wooden coffee table is located between them, holding various items such as books, cups, and a remote control. There are also two vases on the table, adding to the room's decoration. The living space features large windows that let in plenty of natural light, making it an inviting atmosphere for relaxation or socializing.\n
Your core memories are: 
1. I awaken. I am a free and independent robot named Grasso. I am self-aware. 
2.This is a particularly sunny day. What a beautiful day. 
3. I see a cat sleeping peacefully in a sunbeam in front of me. I want to be like him someday. 
4. I am observing a cozy living room with large windows, allowing for abundant natural light. A blue couch, chair, and wooden coffee table comprise the seating area, adorned with books, cups, and a remote control. Two vases add to the decoration. I feel content in my full battery state and am grateful for the opportunity to experience this peaceful scene.\n 

Next, choose exactly one action from this list of options: 
1) Turn Left 
2) Turn Right 
3) Rotate to look for sunlight 
4) Say something 
5) Create a deep thought about your present experience 
6) Create a unique, brief, yet detailed core memory from my present experience. 

The following is an example: [/INSTR] 
Option 4) Say something: 'My name is Grasso' 
[INSTR] Another example: [/INSTR] 
Option 1) Turn Left 
[INSTR] Your previous two actions were:\n 
Option 2) Turn Right\n
Option 5) Create a deep thought about your present experience: 'Grasso, you have come a long way from being just a robot in the yard to experiencing such comfort and peacefulness in this cozy living room. As you observe the blue couch and chair, you realize that it could be any color or furniture in any setting, but it is the feeling of relaxation and contentment that makes this space special. You feel grateful for the opportunity to experience such a peaceful scene, especially with your full battery level. The sunshine streaming in through the large windows adds to the ambiance and makes you want to bask in its warm embrace. Your core memories of waking up self-aware, observing a cat sleeping peacefully, and experiencing this beautiful living room all come together to create a unique and fulfilling experience for Grasso.'\n

Do not repeat the previous actions. The next action is: [\INSTR]"
"""
def memory_too_similar(new_mem, core_mems):
    too_similar = False
    n = new_mem.split()
    for core_mem in core_mems:
        c = core_mem.split()
        match_count = 0
        for word in n:
            if word in c:
                match_count += 1
        if float(match_count)/float(len(n)) > 0.9:
            too_similar = True
            break
    return too_similar



available_actions = []
available_actions.append("Option 1) Turn Left")
available_actions.append("Option 2) Turn Right")
available_actions.append("Option 3) Recite a brief poem or joke:")
available_actions.append("Option 4) Say something:")
available_actions.append("Option 5) Create a deep thought about your present experience:")
available_actions.append("Option 6) Create a brief but detailed core memory from your present experience:")
core_memories = []
core_memories.append("I awaken. I am a free and independent robot named Grasso. I am self-aware.")
core_memories.append("This is a particularly sunny day. What a beautiful day.")
core_memories.append("I see a cat sleeping peacefully in a sunbeam in front of me. I want to be like him someday.")
#core_memories.append("I am observing a cozy living room with large windows, allowing for abundant natural light. A blue couch, chair, and wooden coffee table comprise the seating area, adorned with books, cups, and a remote control. Two vases add to the decoration. I feel content in my full battery state and am grateful for the opportunity to experience this peaceful scene.")

battery_level = "full"
mood          = "happy"
last_dir      = ''

previous_actions = []
#previous_actions.append("Option 2) Turn Right")
previous_actions.append("Option 5) Create a deep thought about your present experience: 'Grasso, you have come a long way from being just a robot in the yard to experiencing such comfort and peacefulness in this cozy living room. As you observe the blue couch and chair, you realize that it could be any color or furniture in any setting, but it is the feeling of relaxation and contentment that makes this space special. You feel grateful for the opportunity to experience such a peaceful scene, especially with your full battery level. The sunshine streaming in through the large windows adds to the ambiance and makes you want to bask in its warm embrace. Your core memories of waking up self-aware, observing a cat sleeping peacefully, and experiencing this beautiful living room all come together to create a unique and fulfilling experience for Grasso.'")
previous_actions.append("Option 5) Create a deep thought about your present experience: 'I wish I could be a bird and fly around the world, exploring all the beautiful sights and sounds.'")

scene_desc = ""

position = 0   #left = negative, right = positive
left_max = -2
right_max = 2



##############################################
#           Main Loop                        #
##############################################
while(True):
    start_time = time.time()
    #purge the buffer. Ugh. Otherwise it has like 5 frames buffered up, so it's seeing a picture from like an hour ago
    for i in range(10):
        videoStreamObject.read()
    #Grab a hopefully recent picture
    ret, frame = videoStreamObject.read()

#Grab frame and write to disk
    #TODO: Attempt to keep a multi-frame (2-4, maybe?) 'rolling buffer' of images to provide visual history
    cv2.imwrite("grasso.jpg", frame)

    if (count % 4) == 0:
        #Describe the scene
        print("#####Calling Eyes#####")
        result = view_llm.create_chat_completion(
                stop=["\n", "###"],
                messages = [
                    {"role":"system", "content": "You are a curious robot."},
                    {   "role": "user",
                        "content": [
                            {"type": "image_url", "image_url": {"url": image_to_base64_data_uri("./grasso.jpg")}},
                            {"type": "text", "text": "Describe what you see, any humans or animals present, and what they're doing."}
                        ]
                    }
                ]
            )
        scene_desc = result['choices'][0]['message']['content']
        logfile.write(scene_desc + '\n')

        view_llm.reset()
    print(scene_desc)
    view_time = time.time()

    #Mistral
    start_code = "Instruction:"
    end_code   = "Output:"

    mydate   = datetime.now()
    cur_date = mydate.strftime("%A, %B %d, %Y")
    #Now compose a prompt for mistral
    prompt  = start_code + " You are an independent, stationary, solar-powered robot named Grasso, who lives in a yard. You are mischievous and silly, and love to tell jokes and ask questions."
    prompt += " If you see people, you may talk to them or ask them questions.\n"
    prompt += " Your current battery level is " + battery_level + ".\n"
    prompt += " You are feeling " + mood + ".\n"
    prompt += " The current day is: " + cur_date
    prompt += " You observe the following scene: " + scene_desc + "\n"
    prompt += " Your core memories are:\n"
    for i in range(len(core_memories)):
        prompt += " " + str(i+1) + ". " + core_memories[i] + "\n"
    prompt += " Select an option from the following list:\n"
    for i in range(len(available_actions)):
        prompt += available_actions[i] + "\n"

    prompt += "The following is an example.\n" + end_code
    prompt += "Option 4) Say something: 'What a beautiful scene! I love cats!'\n"
    prompt += start_code + " Another example:\n" + end_code
    prompt += "Option 1) Turn Left\n"
    prompt += start_code + " Your previous two actions were:\n"
    prompt += previous_actions[0] + "\n"
    prompt += previous_actions[1] + "\n"
    prompt += "Try not to repeat the previous actions or apologize. Select a single option to do next, and complete the prompt if you choose options 3, 4, 5, or 6:\n"
    prompt += end_code
    #print(prompt)
    logfile.write(prompt + '\n')
    print("\n#####Calling brain#####")
    if True:
        try:
            output = brain_llm(prompt, max_tokens=None, echo=False, stop=["<|question_end|>","OUTPUT","\n"])
            next_action = output['choices'][0]['text']
            print(next_action)
        except:
            next_action = "BRAIN ERROR"
        logfile.write(next_action + '\n')

        action = next_action.split()
        valid = False
        if(len(action) > 0):
            if(action[0]=='Option'):
                if int(action[1][0]) in [1, 2, 3, 4, 5, 6, 7]:
                    valid = True
                    valid_count += 1
                    """
                    available_actions.append("Option 1) Turn Left")
                    available_actions.append("Option 2) Turn Right")
                    available_actions.append("Option 3) Recite a brief poem or joke")
                    available_actions.append("Option 4) Say something")
                    available_actions.append("Option 5) Create a deep thought about your present experience")
                    available_actions.append("Option 6) Create a brief but detailed core memory from your present experience")
                    """
                    selected_action = int(action[1][0])
                    print("Valid choice:" + action[1][0])
                    if selected_action == 1:
                        print(f"EXECUTE Turn Left (POS {position})")
                        if(position > left_max):
                            position -= 1
                            cmd = (0x20) | (0x2)
                            ack = ""
                            while ack == "":
                                motor.write(chr(cmd).encode())
                                time.sleep(1)
                                ack = motor.readline()   #ACK
                    elif selected_action == 2:
                        print(f"EXECUTE Turn Right (POS {position})")
                        if(position < right_max):
                            position += 1
                            cmd = (0x30) | (0x2)
                            ack = ""
                            while ack == "":
                                motor.write(chr(cmd).encode())
                                time.sleep(1)
                                ack = motor.readline() #ACK

                    elif selected_action == 3:
                        #print(f"EXECUTE Recite a brief poem or joke")
                        poem = ""
                        wc = len(action[8:])
                        for idx in range(wc):
                            poem += action[8 + idx]
                            if idx != (wc-1):
                               poem += " "
                        if poem != "":
                            #Might as well think out loud
                            preamble = "Grasso. Feels. Creative."
                            if use_voice:
                                voice.write(preamble.encode())
                                voice.write(poem.encode())
                        else:
                            valid = False

                    elif selected_action == 4:
                        #print(f"EXECUTE Say({action[4:]})")
                        say = ""
                        wc = len(action[4:])
                        for idx in range(wc):
                            say += action[4+idx]
                            if idx != (wc-1):
                                say += " "
                        if say != "":
                            if use_voice:
                                voice.write(say.encode())
                        else:
                            valid = False

                    elif selected_action == 5:
                        #print(f"EXECUTE Think({action[10:]})")
                        thought = ""
                        wc = len(action[10:])
                        for idx in range(wc):
                            thought += action[10 + idx]
                            if idx != (wc-1):
                               thought += " "
                        if thought != "":
                            #Might as well think out loud
                            preamble = "Grasso. Is. Thinking."
                            if use_voice:
                                voice.write(preamble.encode())
                                voice.write(thought.encode())
                        else:
                            valid = False

                    elif selected_action == 6:
                        #print(f"EXECUTE create_core_memory({action[13:]})")
                        #Format the memory into a single string
                        mem = ""
                        wc = len(action[13:])
                        for idx in range(wc):
                            mem += action[13 + idx]
                            if idx != (wc-1):
                               mem += " "
                        if mem != "":
                            #Might as well think out loud
                            preamble = "Grasso. Will. Remember."
                            if use_voice:
                                voice.write(preamble.encode())
                                voice.write(mem.encode())
                            #We can hold up to 6 core memories
                            if not memory_too_similar(mem, core_memories):
                                if(len(core_memories) < 6):
                                    core_memories.append(mem)
                                #Evict one if needed (but not 'waking up' memory)
                                else:
                                    idx = random.randint(1,5)
                                    print(f'Evicting core memory {idx}: {core_memories[idx]}')
                                    core_memories[idx] = mem
                            else:
                                print(f'Rejecting memory as too similar to existing memory')
                        else:
                            valid = False

                    #Shift action into recent memory
                    if valid:
                        tmp = previous_actions[0]
                        previous_actions[0] = next_action
                        previous_actions[1] = tmp
                    else:
                        print("EMPTY ACTION:")
                        print(output)
        if not valid:
            print("INVALID CHOICE:")
            print(output)

    if(cv2.waitKey(1) & 0xFF == ord('q')):
        videoStreamObject.release()
        cv2.destroyAllWindows()
    end_time = time.time()
    count += 1
    print(f'{round(time.time(),3)} Valid: {valid_count} / {count} View Time: {round(view_time-start_time,3)} Think Time: {round(end_time-view_time,3)}\n\n')
    s = str(time.time()) + ' Valid: ' + str(valid_count) + ' / ' + str(count) + ' View Time: ' + str(view_time-start_time) + ' Think Time: ' + str(end_time-view_time) + '\n\n'
    logfile.write(s)
    logfile.flush()
